/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * hrtfconv.cpp
 *
 *  Created on: 2011/02/28
 *      Author: HARK Development Team
 */

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <stdexcept>
#include <stdint.h>
#include <unistd.h>
#include <boost/static_assert.hpp>
#include <boost/format.hpp>

#include <libharkio/libharkio.h>

#include "endian.hpp"

using namespace std;
using namespace boost;

#define PRINT(x) do { cout << #x << " = " << x << endl; } while (0)

namespace
{

void PrintUsage()
{
  cout <<
      "Usage: hrtfconv [options] hrtffile harkfile\n"
      "Options:\n"
      "  -h              Converts original format to HARK file format (default)\n"
      "  -H              Converts HARK file format to original format\n"
      "  -c <channel>    Set base channel to <channel>. <channel> must be 'l' or 'r'. Default is 'l'\n"
      ;
}

struct HRTFHeader {
  uint32_t fs;    // sampling rate [Hz]
  uint32_t nfft;  // number of FFT points
  uint32_t tick;  // interval between directions [deg]
  int32_t from;   // start direction [deg] front:0, right:-, left:+
  int32_t to;     // end direction [deg]
  uint32_t fnum;  // number of subbands
  uint32_t dnum;  // number of directions
};

struct HRTF {
  HRTFHeader hh;
  std::vector<double> iid; // IID
  std::vector<double> ipd; // IPD
};

template <typename T>
void Read(istream& is, T& result)
{
  char buf[sizeof(T)];
  is.read(buf, sizeof(T));
  InterpretBytes(result, buf, true); // ファイルの中身はリトルエンディアン
}

void ReadHRTFHeader(istream& is, HRTFHeader& result)
{
  Read(is, result.fs);
  Read(is, result.nfft);
  Read(is, result.tick);
  Read(is, result.from);
  Read(is, result.to);
  Read(is, result.fnum);
  Read(is, result.dnum);
}

void ReadHRTFData(istream& is, std::vector<double>& result)
{
  BOOST_STATIC_ASSERT(sizeof(double) == 8);

  for (size_t i = 0; i < result.size(); ++i)
  {
    Read(is, result[i]);
  }
}

void ReadHRTF(istream& is, HRTF& result)
{
  ReadHRTFHeader(is, result.hh);

  size_t size = result.hh.fnum * result.hh.dnum;

  result.iid.resize(size, 0.0);
  result.ipd.resize(size, 0.0);

  ReadHRTFData(is, result.iid);
  ReadHRTFData(is, result.ipd);
}

void PrintHRTFHeader(const HRTFHeader& h)
{
  cerr << "fs: " << h.fs << endl;    // sampling rate [Hz]
  cerr << "nfft: " << h.nfft << endl;  // number of FFT points
  cerr << "tick: " << h.tick << endl;  // interval between directions [deg]
  cerr << "from: " << h.from << endl;   // start direction [deg] front:0, right:-, left:+
  cerr << "to: " << h.to << endl;     // end direction [deg]
  cerr << "fnum: " << h.fnum << endl;  // number of subbands
  cerr << "dnum: " << h.dnum << endl;  // number of directions
}

void SetMic(harkio::FMT::MIC_LIST_TYPE* micList, float headRadius)
{
  micList->at(0).x = 0;
  micList->at(0).y = -headRadius;
  micList->at(0).z = 0;
  micList->at(1).x = 0;
  micList->at(1).y = headRadius;
  micList->at(1).z = 0;
}

void SetTF(
    const std::vector<double>& iid, const std::vector<double>& ipd,
    size_t begin, size_t fnum,
    std::vector<std::vector<harkio::TF_TYPE> >& result)
{
  if (iid.size() != ipd.size())
  {
    throw length_error("iid.size != ipd.size");
  }
  else if (iid.size() < begin + fnum)
  {
    throw length_error("iid and ipd are too short");
  }
  else if (result.size() != 2)
  {
    throw length_error((format("result.size != 2, actual = %d") % result.size()).str());
  }
  else if (result[0].size() != fnum)
  {
    throw length_error("result[0].size != fnum");
  }
  else if (result[1].size() != fnum)
  {
    throw length_error("result[1].size != fnum");
  }

  /*
   * left channel: 0
   * right channel: 1
   * 左チャンネルを0にして、右チャンネルにデータを入れておく。
   * どうせLocalizeBinauralノードでは、複素数から振幅（および位相）
   * を取り出して両チャンネルの差を取って使うわけだから、問題ない。
   * ただ、どちらのチャンネルが基準側なのかを意識する必要はある。
   * この関数では左チャンネルを基準として固定している。
   */
  for (size_t i = 0; i < fnum; ++i)
  {
    harkio::BASE_TYPE theta = ipd[begin + i];
    harkio::BASE_TYPE r_db = iid[begin + i];
    harkio::BASE_TYPE r = pow(10, r_db / 20);
    //result[0][i] = harkio::TF_TYPE(0, 0);
    //result[1][i] = harkio::TF_TYPE(r * cos(theta), r * sin(theta));
    result[0][i] = polar(static_cast<harkio::BASE_TYPE>(pow(10, 0)), static_cast<harkio::BASE_TYPE>(0));
    result[1][i] = polar(r, theta);
  }
}

void SetTF(
    const std::vector<double>& iid, const std::vector<double>& ipd,
    size_t dnum, size_t fnum,
    std::vector<std::vector<std::vector<std::vector<harkio::TF_TYPE> > > >& result)
{
  if (iid.size() != ipd.size())
  {
    throw length_error("iid.size != ipd.size");
  }
  else if (iid.size() < dnum * fnum)
  {
    throw length_error("iid and ipd are too short");
  }
  else if (result.size() != dnum || result[0].size() != 1 || result[0][0].size() != 2)
  {
    throw length_error("wrong result size");
  }

  for (size_t d = 0; d < dnum; ++d)
  {
    SetTF(iid, ipd, d * fnum, fnum, result[d][0]);
  }
}

void SetTF(
    const std::vector<double>& iid, const std::vector<double>& ipd,
    size_t dnum, size_t fnum,
    std::vector<std::vector<std::vector<harkio::TF_TYPE> > >& result)
{
  if (iid.size() != ipd.size())
  {
    throw length_error("iid.size != ipd.size");
  }
  else if (iid.size() < dnum * fnum)
  {
    throw length_error("iid and ipd are too short");
  }
  else if (result.size() != dnum || result[0].size() != 2)
  {
    throw length_error("wrong result size");
  }

  for (size_t d = 0; d < dnum; ++d)
  {
    SetTF(iid, ipd, d * fnum, fnum, result[d]);
  }
}

void SetSrc(uint32_t dnum, uint32_t tick, int32_t from, double srcDistance, std::vector<std::vector<harkio::MIC_TYPE> >& result)
{
  for (size_t d = 0; d < dnum; ++d)
  {
    int32_t deg = static_cast<int32_t>(d * tick) + from;
    double rad = M_PI * deg / 180.0;
    result[d][0].x = srcDistance * cos(rad);
    result[d][0].y = srcDistance * sin(rad);
    result[d][0].z = 0;
  }
}

void SetSrc(uint32_t dnum, uint32_t tick, int32_t from, double srcDistance, std::vector<harkio::MIC_TYPE>& result)
{
  for (size_t d = 0; d < dnum; ++d)
  {
    int32_t deg = static_cast<int32_t>(d * tick) + from;
    double rad = M_PI * deg / 180.0;
    result[d].x = srcDistance * cos(rad);
    result[d].y = srcDistance * sin(rad);
    result[d].z = 0;
  }
}

void SetSpec(harkio::SPEC& spec)
{
  time_t tim;
  tm* ltim;
  time(&tim);
  ltim = localtime(&tim);

  spec.setCreator("Kazuhiro Nakadai");
  spec.setDate((format("%04d/%02d/%02d") % (ltim->tm_year + 1900) % (ltim->tm_mon + 1) % ltim->tm_mday).str().c_str());
  spec.setMemo("converted by hrtfconv");
}

harkio::HARK* ConvertToM2PG(const HRTF& hrtf)
{
  harkio::HARK* hark = 0;
  harkio::HGTF* hgtf = 0;
  harkio::FMT* fmt = 0;
  harkio::M2PG* m2pg = 0;
  harkio::SPEC* spec = 0;

  try
  {
    hark = new harkio::HARK();
    hgtf = new harkio::HGTF();
    fmt = new harkio::FMT(harkio::FMT::V1_0, harkio::FMT::M2PG, 2, hrtf.hh.nfft, hrtf.hh.nfft / 2 + 1, hrtf.hh.fs);
    m2pg = new harkio::M2PG(harkio::M2PG::V1_0, fmt, hrtf.hh.dnum, 1);
    spec = new harkio::SPEC(harkio::SPEC::V1_0);
  }
  catch (const bad_alloc& e)
  {
    delete hark;
    delete hgtf;
    delete fmt;
    delete m2pg;
    delete spec;
    throw e;
  }

  fmt->newMic();
  SetMic(fmt->getMic(), 0.090); // 頭の半径は9cm
  SetTF(hrtf.iid, hrtf.ipd, hrtf.hh.dnum, hrtf.hh.fnum, *m2pg->getTF());
  SetSrc(hrtf.hh.dnum, hrtf.hh.tick, hrtf.hh.from, 1.0, *m2pg->getSrc());
  SetSpec(*spec);


  hgtf->addChunk(fmt);
  hgtf->addChunk(m2pg);
  hgtf->addChunk(spec);

  hark->addData(hgtf);
  return hark;
}

harkio::HARK* ConvertToGTF(const HRTF& hrtf)
{
  harkio::HARK* hark = 0;
  harkio::HGTF* hgtf = 0;
  harkio::FMT* fmt = 0;
  harkio::GTF* gtf = 0;
  harkio::SPEC* spec = 0;

  try
  {
    hark = new harkio::HARK();
    hgtf = new harkio::HGTF();
    fmt = new harkio::FMT(harkio::FMT::V1_0, harkio::FMT::GTF, 2, hrtf.hh.nfft, hrtf.hh.fnum, hrtf.hh.fs);
    gtf = new harkio::GTF(harkio::GTF::V1_0, fmt, hrtf.hh.dnum, hrtf.hh.fnum);
    spec = new harkio::SPEC(harkio::SPEC::V1_0);
  }
  catch (const bad_alloc& e)
  {
    delete hark;
    delete hgtf;
    delete fmt;
    delete gtf;
    delete spec;
    throw e;
  }

  fmt->newMic();
  SetMic(fmt->getMic(), 0.090); // 頭の半径は9cm
  SetTF(hrtf.iid, hrtf.ipd, hrtf.hh.dnum, hrtf.hh.fnum, *gtf->getTF());
  SetSrc(hrtf.hh.dnum, hrtf.hh.tick, hrtf.hh.from, 1.0, *gtf->getSrc());
  SetSpec(*spec);


  hgtf->addChunk(fmt);
  hgtf->addChunk(gtf);
  hgtf->addChunk(spec);

  hark->addData(hgtf);
  return hark;
}

}

int main(int argc, char **argv)
{
#if 1
  bool toHARKFormat = true;
  bool baseChannelIsLeft = true;

  int ch;
  extern char *optarg;
  extern int optind, opterr;

  while ((ch = getopt(argc, argv, "hHc:")) != -1)
  {
    switch (ch)
    {
    case 'h':
      toHARKFormat = true;
      break;
    case 'H':
      toHARKFormat = false;
      break;
    case 'c':
      switch (optarg[0])
      {
      case 'l': baseChannelIsLeft = true; break;
      case 'r': baseChannelIsLeft = false; break;
      default: baseChannelIsLeft = true;
      }
      break;
    default:
      PrintUsage();
    }
  }
  argc -= optind;
  argv += optind;
  if (argc != 2)
  {
    cerr << "need a hrtf file name and hark file name" << endl;
    PrintUsage();
    exit(1);
  }

  string hrtfFilename(argv[0]);
  string harkFilename(argv[1]);
  ifstream ifs(hrtfFilename.c_str(), ios::in | ios::binary);

  HRTF hrtf;
  ReadHRTF(ifs, hrtf);
  ifs.close();
  PrintHRTFHeader(hrtf.hh);

  cerr << "creating M2PG" << endl;
  harkio::HARK* m2pgFile = ConvertToM2PG(hrtf);
  m2pgFile->writeFile(harkFilename + ".dat");

  cerr << "creating GTF" << endl;
  harkio::HARK* gtfFile = ConvertToGTF(hrtf);
  gtfFile->writeFile(harkFilename + ".tf");

#endif
#if 0
  char buf[4] = { 0x00, 0x00, 0x01, 0x00 };

  uint32_t i = InterpretBytes<uint32_t>(buf, false);
  cout << i << endl;

  i = InterpretBytes<uint32_t>(buf, true);
  cout << i << endl;
#endif


  return 0;
}
