/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Timeval.hpp
 *
 *  Created on: 2011/01/31
 *      Author: HARK Development Team
 */

#ifndef TIMEVAL_HPP_
#define TIMEVAL_HPP_

/**
 * @file Timeval.hpp
 * timeval構造体のための関数・演算子ライブラリ.
 * このライブラリは、従来のtimeradd, timersub, timercmp, timerclear, timerissetマクロの代わりとなる
 * 関数・演算子を提供する。
 * 従来のマクロに比べ、このライブラリで提供する機能はC++の機能を用いて作られており、
 * 安全・便利である。（ただし、従来のマクロを内部的に利用している）
 */

#include <sys/time.h>

/**
 * timeval同士の加算を行う.
 */
inline
timeval operator +(const timeval& lhs, const timeval& rhs)
{
    timeval result;
    timeradd(&lhs, &rhs, &result);
    return result;
}

/**
 * timeval同士の減算を行う.
 */
inline
timeval operator -(const timeval& lhs, const timeval& rhs)
{
    timeval result;
    timersub(&lhs, &rhs, &result);
    return result;
}

/**
 * timeval同士の比較を行う.
 */
inline
bool operator <(const timeval& lhs, const timeval& rhs)
{
    return timercmp(&lhs, &rhs, <);
}

/**
 * timeval同士の比較を行う.
 */
inline
bool operator <=(const timeval& lhs, const timeval& rhs)
{
    return timercmp(&lhs, &rhs, <=);
}

/**
 * timeval同士の比較を行う.
 */
inline
bool operator >(const timeval& lhs, const timeval& rhs)
{
    return timercmp(&lhs, &rhs, >);
}

/**
 * timeval同士の比較を行う.
 */
inline
bool operator >=(const timeval& lhs, const timeval& rhs)
{
    return timercmp(&lhs, &rhs, >=);
}

/**
 * timevalをクリアする.
 */
inline
void TimerClear(timeval& tim)
{
    timerclear(&tim);
}

/**
 * timevalに値が設定されているかどうか.
 */
inline
bool TimerIsSet(const timeval& tim)
{
    return timerisset(&tim) == 0 ? false : true;
}

inline
double TimevalToSeconds(const timeval& tim)
{
    return static_cast<double>(tim.tv_sec) + tim.tv_usec * 1e-6;
}

#endif /* TIMEVAL_HPP_ */
