/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * DirectionJudgeImpl.cpp
 *
 *  Created on: 2011/01/14
 *      Author: HARK Development Team
 */

#include <numeric> // accumulate
#include <algorithm> // transform, fill
#include <functional>
#include <sstream>

#include <boost/function.hpp>
#include <boost/bind.hpp>

#include <DimensionPromise/dimpro.hpp>

#include <DirectionJudgeImpl.hpp>
#include <Epipolar.hpp>
#include <tickdegree.hpp> // TICKDEGREE
#include <Logger.hpp>
#include <Probability.hpp>
#include <Utils.hpp>
#include <IIPDS.hpp>
#include <DirectionConverter.hpp>

using namespace std;
using namespace boost;
using namespace NSDimensionPromise;

namespace NSHarkBinaural
{

/**
 * model で示されるデータに近づけるように input のデータの回り込みを補正する.
 * model をベースとしたキャリブレーション
 *
 * @param input IPD入力
 * @param model IPDのモデル
 */
Angle CalibrateDirection(Angle input, Angle model)
{
    Angle diff = Angles::Zero;
    int ratio;
    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    diff = input - model;
    if (diff < -Angles::AnglePi) {
        ratio = static_cast<int>((diff / 2.0 / Angles::AnglePi - 0.5).X());
        diff = diff - 2.0 * Angles::AnglePi * ratio;
    } else if (diff > Angles::AnglePi) {
        ratio = static_cast<int>((diff / 2.0 / Angles::AnglePi + 0.5).X());
        diff = diff - 2.0 * Angles::AnglePi * ratio;
    }

    LogFinest(logger, format("input is %f rad, model is %f rad, diff is %f rad")
              % input.ToRadian() % model.ToRadian() % diff.ToRadian());

    return diff + model;
}

namespace
{

double CalculateIPDDistance(const vector<Angle>& input, const vector<Angle>& eipd, const vector<BandAndHarmonicRatio>& ipdh)
{
    double d = 0.0;
    double diff;

    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    for (size_t i = 0; i < ipdh.size(); ++i)
    {
        diff = (input.at(i) - eipd.at(i)).X();
        d += diff * diff / ipdh[i].band.Index();
        LogFinest(logger, format("diff is %f, d is %f") % diff % d);
    }
    d /= ipdh.size();
    return d;
}

double CalculateIIDDistance(const IID& iid, const vector<double>& hrtfiid, const vector<BandAndHarmonicRatio>& iidh)
{
    double d = 0.0;
    double diff;

    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    for (size_t i = 0; i < iidh.size(); ++i)
    {
        size_t iidh_i = iidh.at(i).band.Index();
        diff = iid[iidh_i] - hrtfiid[iidh_i];
        d += diff * diff * iidh_i;
        LogFinest(logger, format("diff is %f, d is %f") % diff % d);
    }
    d /= iidh.size();
    return d;
}

#if 0
struct Power2
{
    double operator()(double a, double b) const
    {
        return a + b * b;
    }
};
// 周波数と音路差から、位相差を計算する
struct FreqToPhaseDiff
{
    Length dl;
    FreqToPhaseDiff(Length& dl) : dl(dl) { }
    Angle operator ()(const BandAndHarmonicRatio& bh)
    {
        return PhaseDifference(dl, bh.band.Freq());
    }
};
// IPDをモデルに基づいて調整する
struct PhaseDiffCalibrator
{
    const IPD& ipd;
    PhaseDiffCalibrator(const IPD& ipd) : ipd(ipd) { };

    /**
     * IPDをモデルに基づいて調整する.
     * @param bh 対象の周波数バンド
     * @param phaseDiff 右チャンネル基準のモデルデータ
     */
    Angle operator ()(const BandAndHarmonicRatio& bh, Angle phaseDiff)
    {
        return CalibrateDirection(
                   Angle(ipd.RelativeTo() == Channel::R ? ipd[bh.band.Index()] : -ipd[bh.band.Index()]),
                   phaseDiff);
    }
};
#endif

// 確率の比較
struct GreaterProb
{
    typedef pair<size_t, Probability> value_type;
    bool operator ()(const value_type& a, const value_type& b)
    {
        return a.second.Get() > b.second.Get();
    }
};

/**
 * インデックス0に対応する角度を取得する.
 */
int BeginDegree()
{
    return gDirConv.BeginDegree();
}

/**
 * 最終インデックスに対応する角度を取得する.
 */
int EndDegree()
{
    return -gDirConv.BeginDegree();
}

/**
 * 角度の個数を取得する.
 * 具体的には最終インデックス番号+1を返す
 */
size_t NumberOfDirections()
{
    return gDirConv.ToIndex(EndDegree()) + 1;
}

/**
 *
 * 距離が最小値を持つものが答えである確信度を正規分布を仮定し確率密度関数から求める.
 *
 * 正規化した確率密度関数
 *   1 / sqrt(2 * pi) * exp( -x * x / 2 )
 * の [0,x] の不定積分は誤差関数を使って
 *   1 / 2 * erf( x / sqrt(2) )
 * とあらわせる。
 * これを [ -inf, x ] の積分値に変換して確信度とする
 *
 */
vector<Probability> CalculateProbability(const vector<double>& distance)
{
    /*
    struct Power2: public binary_function<double, double, double>
    {
      double operator ()(const double& init, const double& val) const
      { return init + val * val; }
    };
    */

    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    vector<Probability> prob(distance.size());
    const double m = accumulate(distance.begin(), distance.end(), 0.0) / distance.size();
    //const double v = accumulate(distance.begin(), distance.end(), 0.0, Power2()) / distance.size() - m * m;

    double v = 0;
    for (vector<double>::const_iterator it = distance.begin(); it != distance.end(); ++it)
    {
        v += pow(*it, 2);
    }
    v -= m * m;

    const double sqrt_v = sqrt(v);
    double x;

    LogFinest(logger, format("m = %f, v = %f") % m % v);

    if (fabs(v) < 1e-9)
    {
        fill(prob.begin(), prob.end(), 0.5);
    }
    else
    {
        for (size_t i = 0; i < prob.size(); ++i)
        {
#if 0
            x = fabs(m - distance.at(i)) / sqrt_v;
            prob[i] = erf(x / M_SQRT2) / 2;
            if (distance[i] > m)
            {
                prob[i] = Probability(0.5 - prob[i].Get());
            }
            else
            {
                prob[i] = Probability(0.5 + prob[i].Get());
            }
#else
            x = (m - distance.at(i)) / sqrt_v;
            try
            {
                prob[i] = erf(x / M_SQRT2) / 2 + 0.5;
            }
            catch (out_of_range& exp)
            {
                LogSevere(logger, exp.what());
                throw;
            }
#endif
        }
    }

    return prob;
}

/**
 * Demplster結合則により、2つの値を結合する.
 */
double Dempster( double a, double b ) {
    double d;
    d = a * b + a * (1 - b) + (1 - a) * b;
    return d;
}

/**
 * 音源からの左右の耳の距離差モデルを求める.
 * 右チャンネル基準での相対距離差のモデルを求める。
 * ここでいうモデルとは、ある角度における左右の距離差の計算上の値のこと。
 * @return 左右の距離差モデル（右チャンネル基準の相対距離）
 */
vector<Length> CalcLRLengthDifference(const Epipolar& epipolar, int method)
{
    //const int middleDir = 180 / TICKDEGREE / 2;
    const int middleDir = gDirConv.ToIndex(0); // 正面（0度）に対応するインデックス番号を取得
    //vector<Length> dl(180 / TICKDEGREE + 1);
    vector<Length> dl(NumberOfDirections());
    function<Length (Angle)> lrDifferenceMethod;
    if (method == 0)
    {
        // &epipolarをEpipolar::LRDifferenceSimpleの第一引数に束縛する
        // lrDifferenceMethodの第一引数をEpipolar::LRDifferenceSimpleの第二引数に束縛する
        lrDifferenceMethod =
            bind(&Epipolar::LRDifferenceSimple, &epipolar, _1);
    }
    else if (method == 1)
    {
        lrDifferenceMethod =
            bind(&Epipolar::LRDifferenceRevised, &epipolar, _1);
    }
    if (lrDifferenceMethod)
    {
        // まず正面を計算（0になるはずだけど、一応真面目に計算）
        dl[middleDir] = lrDifferenceMethod(Angle::FromDegree(0));
        // 正面以外を計算
        for (size_t i = 0; i < middleDir; ++i)
        {
            dl[i] = lrDifferenceMethod(Angle::FromDegree(gDirConv.ToDegree(i)));
            dl[dl.size() - 1 - i] = -dl[i];
        }

        // dlは、左に行くほど負の無限大, 右に行くほど正の無限大に近づく
        // 「近づく」というのは、そちらの方向に動くということで、
        // とても大きな数字になるわけではない。
    }
    return dl;
}


} // namespace

JudgeEpipolar::JudgeEpipolar(NSDimensionPromise::Length sourceDistance, NSDimensionPromise::Length headRadius)
    : sourceDistance(sourceDistance), headRadius(headRadius), lrLengthDifference()
{

}

JudgeEpipolar::DirectionProbabilityVectorPtr
JudgeEpipolar::Judge(const std::pair<BandHarmonicVectorPtr, BandHarmonicVectorPtr>& dividedBands,
                     IPD::ptr_t ipd, IID::ptr_t iid, MethodFlag::Kind mflag, bool sortByProbability)
{
    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    ostringstream logOss; // ログメッセージ構築用

    DirectionProbabilityVectorPtr result(new DirectionProbabilityVector());

    // 信号源からの距離差計算用
    Epipolar epipolar(sourceDistance, headRadius);

#if defined(SIMPLEEPIPOLAR)
    const int METHOD = 0;
#else
    const int METHOD = 1;
#endif

    //if (this->lrLengthDifference.size() != 180 / TICKDEGREE + 1)
    if (this->lrLengthDifference.size() != NumberOfDirections())
    {
        LogFine(logger, "calculating dl");
        this->lrLengthDifference = CalcLRLengthDifference(epipolar, METHOD);
        if (logger.GetLevel().level <= LogLevel::FINEST)
        {
            logOss.flush();
            logOss << "got dl\n";
            for (size_t i = 0; i < this->lrLengthDifference.size(); ++i)
            {
                logOss << (boost::format("%f: ") % this->lrLengthDifference[i].X()).str();
            }
            LogFinest(logger, logOss.str());
        }
    }

    BandHarmonicVectorPtr ipdh(dividedBands.first); // 低周波数帯
    BandHarmonicVectorPtr iidh(dividedBands.second); // 高周波数帯

    vector<Probability> pipd, piid, pdempster;
    vector<pair<size_t, Probability> > pmax;

    // IPDを用いて計算
    if (ipdh) LogFine(logger, format("ipdh size: %d") % ipdh->size());
    else LogInfo(logger, "ipdh is null");

    if (ipdh && ipdh->size() > 0 && mflag != MethodFlag::IIDONLY)
    {
        pipd = CalcProbabilityFromIPD(ipdh, ipd);
    }

    // IIDを用いて計算
    if (iidh) LogFine(logger, format("iidh size: %d") % iidh->size());
    else LogInfo(logger, "iidh is null");

    if (iidh && iidh->size() > 0 && mflag != MethodFlag::IPDONLY)
    {
        piid = CalcProbabilityFromIID(iidh, iid);
    }

    // IPD, IID両方を用いて計算
    if (mflag == MethodFlag::BOTH)
    {
        if (ipdh->size() > 0 && iidh->size() > 0)
        {
            pdempster.resize(pipd.size());

            // Dempster結合則より統合
            // Dempster(pipd, piid) => pdempster
            transform(pipd.begin(), pipd.end(), piid.begin(), pdempster.begin(), Dempster);
        }
        else if (ipdh->size() == 0 && iidh->size() > 0)
        {
            pdempster = piid;
        }
        else if (ipdh->size() > 0 && iidh->size() == 0)
        {
            pdempster = pipd;
        }
        else
        {
            pdempster.clear();
        }
    }

    // 確率配列の各要素に、その確率が対応している方向を表すインデックスをくっつける。
    // ソートしても、どの方向に対する確率なのか分かるようにするため。
    switch (mflag)
    {
    case MethodFlag::IPDONLY:
        pmax = MakeIndexedVector(pipd);
        break;
    case MethodFlag::IIDONLY:
        pmax = MakeIndexedVector(piid);
        break;
    case MethodFlag::BOTH:
        pmax = MakeIndexedVector(pdempster);
        break;
    default:
        break;
    }
    if (sortByProbability)
    {
        sort(pmax.begin(), pmax.end(), GreaterProb());
    }

    // (index, prob)の組を(angle, prob)の組へ変換
    /*
    for (vector<pair<size_t, Probability> >::iterator it = pmax.begin(); it != pmax.end(); ++it)
    {
      AngleAndProbability aap = MakeAngleAndProbability(*it);
      LogFinest(logger, format("pushing back: <%d = %f deg, %f prob>") % it->first % aap.angle.ToDegree() % aap.probability.Get());
      result->push_back(aap);
    }
    */
    *result = pmax;

    if (result->size() == 0)
    {
        LogWarning(logger, format("result size is zero, mflag == %d") % (int)mflag);
    }

    return result;
}

std::vector<Probability> JudgeEpipolar::CalcProbabilityFromIPD(const BandHarmonicVectorPtr ipdh, IPD::ptr_t ipd)
{
    vector<Probability> pipd;
    if (ipdh && ipdh->size() > 0)
    {
        Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
        ostringstream logOss;

        // TICKDEGREE == 5なら、次の37種類の角度について計算する
        // +90 +85, +80, ... , +5, 0, -5, ... , -80, -85, -90
        // 左耳が+90度、右耳が-90度に該当する
        //vector<double> dipd(180 / TICKDEGREE + 1); // 距離
        vector<double> dipd(NumberOfDirections()); // 距離

        vector<Angle> eipd(ipdh->size()); // モデルデータ
        vector<Angle> newinput(ipdh->size()); // 補正後入力データ

        vector<Length>& dl = this->lrLengthDifference; // 各角度における両耳間音路差

        logOss.str("");
        logOss << "phase difference: ";
        for (int f = 100; f <= 1000; f += 100)
        {
            logOss << format("%lf, ") % PhaseDifference(Length::FromMeter(f), Frequency::FromHz(f)).ToRadian() << endl;
        }
        LogFinest(logger, logOss.str());

        for (size_t dir = 0; dir < dipd.size(); ++dir)
        {
            // 調波構造の各周波数について、両耳間位相差を計算する
            for (size_t j = 0; j < ipdh->size(); ++j)
            {
                eipd[j] = PhaseDifference(dl[dir], (*ipdh)[j].band.Freq());
            }
            LogFine(logger, format("freq to phase diff  dl %f m, ipd %f Hz, model %f deg") %
                    dl[0].ToMeter() % ipdh->at(0).band.Freq().ToHz() % eipd[0].ToDegree());


            // この時点で、dl[dir]の符号とeipdの各要素の符号は一致している
            // eipdは右チャンネル基準のモデルデータ（実際の入力信号に関係ない値）

            // 符号反転すると、基準となる側（左または右）が入れ替わる
            //transform(eipd.begin(), eipd.end(), eipd.begin(), bind1st(minus<Angle>(), Angles::Zero));

            // 入力信号をモデルになるべく近づくように補正
            // PhaseDiffCalibrator( ipdh, eipd ) => newinput
            //transform(ipdh->begin(), ipdh->end(), eipd.begin(), newinput.begin(), PhaseDiffCalibrator(*ipd));
            for (size_t j = 0; j < ipdh->size(); ++j)
            {
                float input = ipd->At((*ipdh)[j].band.Index());
                input = ipd->RelativeTo() == Channel::R ? input : -input;
                newinput[j] = CalibrateDirection(Angle::FromRadian(input), eipd[j]);
            }
            // newinputは右チャンネル基準

            logOss.str("");
            logOss << "got eipd: ";
            for (int j = 0; j < min(eipd.size(), (size_t)3); ++j)
            {
                logOss << eipd[j].ToRadian() << ", ";
            }
            LogFinest(logger, logOss.str());

            logOss.str("");
            logOss << "ipdh: ";
            for (int j = 0; j < min(ipdh->size(), (size_t)3); ++j)
            {
                logOss << ipdh->at(j).band.Freq().ToHz() << ", ";
            }
            LogFinest(logger, logOss.str());

            logOss.str("");
            logOss << "test phase diff: ";
            for (int j = 0; j < min(ipdh->size(), (size_t)3); ++j)
            {
                logOss << format("PhaseDifference(%f, %f).ToRadian() = ")
                    % dl[dir].ToMeter() % ipdh->at(j).band.Freq().ToHz()
                    << " " <<  PhaseDifference(dl[dir], ipdh->at(j).band.Freq()).ToRadian() << ", ";
            }
            LogFinest(logger, logOss.str());

            logOss.str("");
            logOss << "got newinput: ";
            for (int j = 0; j < min(newinput.size(), (size_t)3); ++j)
            {
                logOss << newinput[j].ToRadian() << ", ";
            }
            LogFinest(logger, logOss.str());

            // 位相差からi方向の距離を求める
            // 入力信号がモデルと一致するなら、この距離は0になる
            dipd[dir] = CalculateIPDDistance(newinput, eipd, *ipdh);

            LogFinest(logger, format("dipd[%d] is %f") % dir % dipd[dir]);
        }

        // 距離が小さいほど、答え（モデルデータ）に近いと考えられる
        pipd = CalculateProbability(dipd);
    }
    return pipd;
}

std::vector<Probability> JudgeEpipolar::CalcProbabilityFromIID(const BandHarmonicVectorPtr iidh, IID::ptr_t iid)
{
    vector<Probability> piid;
    if (iidh && iidh->size() > 0)
    {
        Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
        ostringstream logOss;

        double iidsum = 0.0;
        for (size_t i = 0; i < iidh->size(); ++i)
        {
            iidsum += iid->At((*iidh)[i].band.Index());
        }

        // 強度差から確率を与える(適当)

        const size_t dirCount = NumberOfDirections();
        vector<Probability> piidr(dirCount);
        vector<Probability> piidl(dirCount);

        // set each piidr[i]
        fill(piidr.begin(), piidr.begin() + dirCount / 3 + 1, 0.35);
        fill(piidr.begin() + dirCount / 3 + 1, piidr.begin() + dirCount * 2 / 3, 0.5);
        fill(piidr.begin() + dirCount * 2 / 3, piidr.end(), 0.65);

        // piidl[i] = 1.0 - piidr[i]
        transform(piidr.begin(), piidr.end(), piidl.begin(), bind1st(minus<double>(), 1.0));

        if (logger.GetLevel().level <= LogLevel::FINEST)
        {
            logOss.flush();
            logOss << "got piidr and piidl\n";
            for (size_t i = 0; i < piidr.size(); ++i)
            {
                logOss << boost::format("%f %f: ") % piidr[i].Get() % piidl[i].Get();
            }
            LogFinest(logger, logOss.str());
        }

        // 両耳間強度差で、右と左のうち値が大きな方を信用する
        if ((iid->RelativeTo() == Channel::R && iidsum < 0) || (iid->RelativeTo() == Channel::L && iidsum >= 0))
        {
            piid.swap(piidr);
        }
        else
        {
            piid.swap(piidl);
        }
    }
    return piid;
}

/*
AngleAndProbability JudgeEpipolar::MakeAngleAndProbability(const std::pair<size_t, Probability>& indexedProbability)
{
  Angle direction = Angle::FromDegree(90 - static_cast<signed long>(indexedProbability.first) * TICKDEGREE);
  return AngleAndProbability(direction, indexedProbability.second);
}
*/

JudgeHRTFMeasured::JudgeHRTFMeasured(boost::shared_ptr<IIPDS> iipds)
    : iipds(iipds)
{

}

DirectionJudge::DirectionProbabilityVectorPtr
JudgeHRTFMeasured::Judge(const std::pair<BandHarmonicVectorPtr, BandHarmonicVectorPtr>& dividedBands,
                         IPD::ptr_t ipd, IID::ptr_t iid, MethodFlag::Kind mflag, bool sortByProbability)
{
    BandHarmonicVectorPtr ipdh(dividedBands.first); // 低周波数帯
    BandHarmonicVectorPtr iidh(dividedBands.second); // 高周波数帯

    vector<Angle> eipd(ipdh->size());
    vector<Angle> newinput(ipdh->size());
    vector<double> dipd(iipds->ndir, 0); // 距離
    vector<double> diid(iipds->ndir, 0); // 距離
    vector<Probability> pipd, piid, pdempster;
    vector<pair<size_t, Probability> > pmax;
    //AngleProbabilityVectorPtr result(new AngleProbabilityVector());
    DirectionProbabilityVectorPtr result(new DirectionProbabilityVector());

    for (size_t i = 0; i < iipds->ndir; ++i)
    {
        IIPD& iipd_ipd = iipds->ipd[i];
        for (size_t j = 0; j < ipdh->size(); ++j)
        {
            size_t ipdh_i = ipdh->at(j).band.Index();
            newinput[j] = CalibrateDirection(Angle::FromRadian(ipd->At(ipdh_i)), Angle::FromRadian(iipd_ipd.m.at(ipdh_i)));
            eipd[j] = Angle::FromRadian(iipd_ipd.m.at(ipdh_i));
        }

        if (ipdh->size() != 0)
        {
            // 位相差からi方向の距離を求める
            // 入力信号がモデルと一致するなら、この距離は0になる
            dipd[i] = CalculateIPDDistance(newinput, eipd, *ipdh);
        }

        if (iidh->size() != 0)
        {
            IIPD& iipd_iid = iipds->iid[i];
            // 位相差からi方向の距離を求める
            // 入力信号がモデルと一致するなら、この距離は0になる
            diid[i] = CalculateIIDDistance(*iid, iipd_iid.m, *iidh);
        }
    }

    if (ipdh->size() > 0 && mflag != MethodFlag::IIDONLY)
    {
        pipd = CalculateProbability(dipd);
    }
    if (iidh->size() > 0 && mflag != MethodFlag::IPDONLY)
    {
        piid = CalculateProbability(diid);
    }
    // IPD, IID両方を用いて計算
    if (mflag == MethodFlag::BOTH)
    {
        if (ipdh->size() > 0 && iidh->size() > 0)
        {
            pdempster.resize(pipd.size());

            // Dempster結合則より統合
            // Dempster(pipd, piid) => pdempster
            transform(pipd.begin(), pipd.end(), piid.begin(), pdempster.begin(), Dempster);
        }
        else if (ipdh->size() == 0 && iidh->size() > 0)
        {
            pdempster = piid;
        }
        else if (ipdh->size() > 0 && iidh->size() == 0)
        {
            pdempster = pipd;
        }
        else
        {
            pdempster.clear();
        }
    }

    // 確率配列の各要素に、その確率が対応している方向を表すインデックスをくっつける。
    // ソートしても、どの方向に対する確率なのか分かるようにするため。
    switch (mflag)
    {
    case MethodFlag::IPDONLY:
        pmax = MakeIndexedVector(pipd);
        break;
    case MethodFlag::IIDONLY:
        pmax = MakeIndexedVector(piid);
        break;
    case MethodFlag::BOTH:
        pmax = MakeIndexedVector(pdempster);
        break;
    default:
        break;
    }
    if (sortByProbability)
    {
        sort(pmax.begin(), pmax.end(), GreaterProb());
    }

    // (index, prob)の組を(angle, prob)の組へ変換
    /*
    for (vector<pair<size_t, Probability> >::iterator it = pmax.begin(); it != pmax.end(); ++it)
    {
      AngleAndProbability aap = MakeAngleAndProbability(*it);
    //    LogFinest(logger, format("pushing back: <%d = %f deg, %f prob>") % it->first % aap.angle.ToDegree() % aap.probability.Get());
      result->push_back(aap);
    }
    */

    if (result->size() == 0)
    {
//    LogWarning(logger, format("result size is zero, mflag == %d") % (int)mflag);
    }

    result->swap(pmax);
    return result;
}

/*
AngleAndProbability
JudgeHRTFMeasured::MakeAngleAndProbability(const std::pair<size_t, Probability>& indexedProbability)
{
    Angle direction = Angle::FromDegree(-90 + static_cast<signed long>(indexedProbability.first) * iipds->dirtick);
    return AngleAndProbability(direction, indexedProbability.second);
}
*/

}

